/*============================================================================
                HEAP SORT with heap and pointers
   node = Values  |  Values x Values  |  node x Values x node
        (leafnode)     (unarynode)           (binarynode)
To see how the heap is modified during sorting, one should have: history=true.
=========================================================================== */
#include <iostream.h>
#include <stdlib.h>   // file needed for the EXIT_SUCCESS and the EXIT_FAILURE
//----------------------------------------------------------------------------
//            TYPES
typedef int Values;
enum Tag {leafnode, unarynode, binarynode};
struct node {
       struct Leaf   {Values value;};
       struct Unary  {Values value1; Values son;};  //not: node* son
       struct Binary {Values value2; node* left; node* right;};
       union {Leaf l; Unary u; Binary b;};
       Tag tag;
            };
typedef node* nodeptr;
//----------------------------------------------------------------------------
//            printing a heap
void printtree(node* t)
{if (t==NULL) {cout << "empty heap ";} else
{switch (t->tag){
case leafnode: cout << t->l.value; break;
case unarynode:
        cout << "(" << t->u.value1 << "." << t->u.son << ")"; break;
case binarynode:
        cout << "("; printtree(t->b.left);
        cout << "." << t->b.value2 << ".";
                      printtree(t->b.right);
        cout << ")"; break;
default: cout << " Error 1 in printtree!";
         exit(EXIT_FAILURE);}};}
//----------------------------------------------------------------------------
//            printing an array A from A[0] to A[size-1]
void printar(Values A[], int size)
{for (int i=0; i<size; i=i+1) {cout << A[i] << " ";};}
//----------------------------------------------------------------------------
//            maximum of two Values
Values max(Values a, Values b) {if (a>=b) {return a;} else {return b;}}
//----------------------------------------------------------------------------
//            integer part of log base b of n
int logb(int n)
{int templog=0; int temp=1;
if (n<1) {cout << "Error in logb!"; exit(EXIT_FAILURE);}
else {while (temp<=n) {temp=temp*2; templog=templog+1;};
      return templog-1;};}
//----------------------------------------------------------------------------
//            2 to the power of n
int power_two(int n) {if (n==0) {return 1;} else {return 2*power_two(n-1);};}
//----------------------------------------------------------------------------
//            depth of a heap: depth of leaf=1, depth of unary=2
int depth(node* t)
{switch (t->tag){
case leafnode:   return 1; break;
case unarynode:  return 2; break;
case binarynode: return max(depth(t->b.left),depth(t->b.right))+1; break;
default: cout << "Error in depth!"; exit(EXIT_FAILURE);}}
//----------------------------------------------------------------------------
//            MANIPULATING THE HEAP
//            from an array A TO a heap-scheleton H with values
void vector_to_heap(nodeptr &H, Values A[], int i, int elems)
{int elem1; int elem2;
if (elems==0) {cout <<"Error: empty heap from empty array!";
              exit(EXIT_FAILURE);}
else if (elems==1) {H->tag=leafnode; H->l.value=A[i];}
else if (elems==2) {H->tag=unarynode; H->u.value1=A[i];
                    H->u.son=A[2*i+1];}  //not: 2*i
else {//-------------------- computing the elements of the left and right sons
           int temp=power_two(logb(elems+1));
           if   (temp==elems+1) {elem1=((elems+1)/2)-1; elem2=elem1;}
           else {if ((((3*temp)/2)-1)<=elems)
                {elem1=temp-1; elem2=elems-temp;}
                else {elem1=elems-(temp/2); elem2=(temp/2)-1;};};
      //----------------------------------------------------------------------
          H->tag=binarynode;  H->b.value2=A[i];
          nodeptr H1=new node; nodeptr H2=new node;
          vector_to_heap(H1,A,2*i+1,elem1); //not: 2*i
          vector_to_heap(H2,A,2*i+2,elem2); //not: 2*i+1
          H->b.left=H1; H->b.right=H2;};}
//----------------------------------------------------------------------------
//            DESTRUCTION:  HEAP  ->  (last_ELEM x HEAP)
void destr(nodeptr &H, Values &elem)
{Values t; Values t1;
if (H==NULL) {cout << "Error 1 in destr!";} else
{switch (H->tag){
case  leafnode:  elem=H->l.value; H=NULL; break;
case  unarynode: elem=H->u.son; t=H->u.value1; H->tag=leafnode;
                 H->l.value=t; break;
case  binarynode:if (depth(H->b.left) > depth(H->b.right))
              {destr(H->b.left,elem);}
              else //------------------- depth(H->b.left) == depth(H->b.right)
              {if (depth(H)==2)
               {elem=(H->b.right)->l.value;
               t=H->b.value2; t1=(H->b.left)->l.value;
               H->tag=unarynode; H->u.value1=t;
               H->u.son=t1;}
              else destr(H->b.right,elem);
              }; break;
default: cout << "Error 2 in destr!"; exit(EXIT_FAILURE);};};}
//----------------------------------------------------------------------------
//            TOP: HEAP->integer
Values top(nodeptr &H)
{switch (H->tag){
case  leafnode:   return H->l.value; break;
case  unarynode:  return H->u.value1; break;
case  binarynode: return H->b.value2;};}
//----------------------------------------------------------------------------
//            TOP_UPDATE: (HEAP x newTOP) -> new HEAP
void top_update(nodeptr &H, Values v)
{switch (H->tag){
case  leafnode:   H->l.value=v; break;
case  unarynode:  H->u.value1=v; break;
case  binarynode: H->b.value2=v;};}
//----------------------------------------------------------------------------
//            PERCOLATE_OVER: HEAP -> new_HEAP
void percolate_over(nodeptr &H)
{switch (H->tag){
case  unarynode: if (H->u.value1 < H->u.son)
                 {Values t=H->u.value1; H->u.value1=H->u.son;
                  H->u.son=t;}; break;
case  binarynode: if (H->b.value2 < max(top(H->b.left),top(H->b.right)))
               {if (top(H->b.left) > top(H->b.right))
                 {Values t=top(H->b.left); Values t1=top(H);
                 top_update(H,t); top_update(H->b.left,t1);
                 percolate_over(H->b.left);}
                else {Values t=top(H->b.right); Values t1=H->b.value2;
                     top_update(H,t); top_update(H->b.right,t1);
                     percolate_over(H->b.right);};};};}
//----------------------------------------------------------------------------
//            MAX_AWAY: HEAP -> (new_HEAP x top_of_HEAP)
//            Let H be a heap, not a heap-scheleton.
void max_away(nodeptr &H, Values &m)
{m=top(H); Values t;
switch (H->tag){
case  leafnode:   H=NULL; break;
case  unarynode:  t=H->u.son; H->tag=leafnode; H->l.value=t; break;
case  binarynode: destr(H,t); top_update(H,t);
                  percolate_over(H);};}
//----------------------------------------------------------------------------
//            FIX HEAP
void fix_heap(nodeptr &H)
{switch (H->tag){
case leafnode:   break;
case unarynode:  if (H->u.value1 < H->u.son)
                 {Values t=H->u.value1; H->u.value1=H->u.son;
                  H->u.son=t;}; break;
case binarynode: fix_heap(H->b.left); fix_heap(H->b.right);
                 percolate_over(H); break;
default: cout << "Error in fix_heap!"; exit(EXIT_FAILURE);};}
//----------------------------------------------------------------------------
void main()
{const int dim=6;                    // dim should be >=1
 Values A[dim]={19,20,37,43,76,20};  // given array
 Values B[dim];                      // sorted array
 nodeptr HA=new node;
 Values maxelem=0; bool history=true;
cout  << "\nThe given array is:\n"; printar(A,dim);
if (dim > 1) {
vector_to_heap(HA,A,0,dim);
cout  << "\n\nInitial heap scheleton with values:\n  ";
printtree(HA);
cout  << "\n\nThe initial heap is:\n  "; fix_heap(HA);
printtree(HA); cout << "\n";
for (int i=0; i<dim; i=i+1)
  {max_away(HA,maxelem); if (history==true)
          {cout <<"\n step "<< i+1 << ":   -> max: " << maxelem;
           cout <<"\n         new_heap: "; printtree(HA);}
   B[i]=maxelem;};
cout  << "\n\n==> The sorted array (in descending order) is:\n    ";
printar(B,dim); cout << "\n";}
else {cout  << "\n\n==> The sorted array (in descending order) is:\n    ";
      printar(A,dim); cout << "\n";};
exit(EXIT_SUCCESS);
}
/* input:   output: (for dim=6, history=true)
------------------------------------------------------------------------------
            The given array is:
            19 20 37 43 76 20

            Initial heap scheleton with values:
              ((43.20.76).19.(37.20))

            The initial heap is:
              ((19.43.20).76.(37.20))

             step 0:   -> max: 76
                     new_heap: ((19.20.20).43.37)
             step 1:   -> max: 43
                     new_heap: ((20.19).37.20)
             step 2:   -> max: 37
                     new_heap: (20.20.19)
             step 3:   -> max: 20
                     new_heap: (20.19)
             step 4:   -> max: 20
                     new_heap: 19
             step 5:   -> max: 19
                     new_heap: empty heap

            ==> The sorted array (in descending order) is:
                76 43 37 20 20 19
--------------------------------------------------------------------------- */